/*
 * Copyright (c) 2002 Ralf Habacker  <Ralf.Habacker@freenet.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Id:$
 */

#include <stdio.h>
#include <stdlib.h>

#include "objectfile.h"


// read a dll into the cache 

#ifdef __CYGWIN__
#include <sys/cygwin.h> 
#endif 

#ifdef __CYGWIN__

static char * Win32Path(char * s)
{
  char buf[MAX_PATH];
	if (!s || *s == '\0')
		return ""; 
  cygwin_conv_to_win32_path(s, buf);
  return strdup(buf); 
}
#else 
#define Win32Path(s)  s 
#endif


//------- class ObjectFile ------------------------------------------

ObjectFile::ObjectFile(char *aFileName, bool writeable) 
{
	sections = 0;
	isWritable = writeable;
	FileName = 0; 
	lpFileBase = 0; 
	hfile = 0; 
	hfilemapping = 0; 
	
	// FIXME: add search path 
	hfile = CreateFile(Win32Path(aFileName), writeable ? GENERIC_READ | GENERIC_WRITE : GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL,0);
	if (hfile == INVALID_HANDLE_VALUE) {
		char *basename = strrchr(aFileName,'/');
		basename = basename ? basename+1 : aFileName;
	
		char *s = getenv("PATH"); 
		char path[1024]; 
		strcpy(path,s);
		s = strtok(path,":");
		
		while (s) {
			char name[1024]; 
			strcpy(name,s);
			strcat(name,"/"); 
			strcat(name,basename);
			printf("%s: name: %s\n", __FUNCTION__, name);
			hfile = CreateFile(Win32Path(name), writeable ? GENERIC_READ | GENERIC_WRITE : GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL,0);
				if (hfile != INVALID_HANDLE_VALUE)
					break; 
			s = strtok(NULL,":");		
		}
		if (!s) {
			Error = 2;
			return; 
		}
	}
	
	hfilemapping = CreateFileMapping(hfile, NULL, writeable ? PAGE_READWRITE : PAGE_READONLY , 0, 0,  NULL);
	if (hfilemapping == 0) {
		CloseHandle(hfile);
		Error = 2;
		return; 
	}
	
	lpFileBase = MapViewOfFile(hfilemapping, writeable ? FILE_MAP_WRITE : FILE_MAP_READ,0, 0, 0);
	if (lpFileBase == 0) {
		CloseHandle(hfilemapping);
		CloseHandle(hfile);
		Error = 3;
		return; 
	}
	
	// create shortcuts 
	PIMAGE_DOS_HEADER dosheader = (PIMAGE_DOS_HEADER)lpFileBase; 
	
	ntheader = (PIMAGE_NT_HEADERS) ((char *)dosheader + dosheader->e_lfanew);

	if (ntheader->Signature != 0x00004550) {
		Error = 4;
		return; 
	}
	
	sections = new SectionList(lpFileBase); 
	
	ImageBase = ntheader->OptionalHeader.ImageBase;
	FileName = strdup(aFileName); 
	
	Error = 0;
}

ObjectFile::~ObjectFile() 
{
	if (sections) delete sections; 
	if (FileName) free(FileName);
	if (lpFileBase) UnmapViewOfFile(lpFileBase);
	if (hfilemapping) CloseHandle(hfilemapping);
	if (hfile) CloseHandle(hfile);
}






LinkedObjectFile::LinkedObjectFile(char *aFileName, bool writable) : ObjectFile(aFileName,writable) 
{
	exports = 0; 
	imports = 0; 
	relocs = 0; 
	
	if (Error)
		return; 

	if (debug) {
		fprintf(stderr,"Base:       0x%08x\n",lpFileBase);	
		fprintf(stderr,"ImageBase:  0x%08x\n",ImageBase);	
	}

	Section *edata = sections->find(".edata"); 
	if (edata)
		exports = new Exports(*edata);
	else 
		exports = new Exports(*sections,&ntheader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT]);


	Section *idata = sections->find(".idata"); 
	if (idata)
		imports = new Imports(*idata);
	else 
		imports = new Imports(*sections,&ntheader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT]);

	

	relocs = new Relocations(*sections,".reloc");

}

bool LinkedObjectFile::rebind(ObjectFileList &cache) 
{
	PIMAGE_IMPORT_DESCRIPTOR p;

	Section *edata = sections->find(".edata"); 
	Section *idata = sections->find(".idata"); 
	Section *text = sections->find(".text"); 

	LinkedObjectFile *obj; 

	imports->reset(); 
		
	while (p = imports->getNextDescriptor()) {
		bool autoImportFlag; 
		int *patch_address; 
		char *dllname = (char *)p->Name + idata->getAdjust(); 
		fprintf(stderr,"%s - %s \n",__FUNCTION__, dllname);

		if (!(obj = (LinkedObjectFile *)cache.get(dllname) ) ) {
			obj = new LinkedObjectFile(dllname);
			if (obj->getError()) {
				fprintf(stderr,"cant load dll '%s'\n",dllname);
				delete obj; 
				continue; 
			}
			cache.add(obj); 
		}
		PIMAGE_THUNK_DATA hintArray = PIMAGE_THUNK_DATA ((uint) p->OriginalFirstThunk + imports->getAdjust());

		PIMAGE_THUNK_DATA firstArray; 
		fprintf(stderr,"FirstThunk %08x\n",p->FirstThunk);

		if ((autoImportFlag = text->isIn((uint)p->FirstThunk)))
			firstArray = PIMAGE_THUNK_DATA ((uint) p->FirstThunk + text->getAdjust());
		else 
			firstArray = PIMAGE_THUNK_DATA ((uint) p->FirstThunk + imports->getAdjust());

		fprintf(stderr,"FirstArray %08x\n",firstArray);

		for (; hintArray->u1.Function; hintArray++, firstArray++) {
			PIMAGE_IMPORT_BY_NAME a = PIMAGE_IMPORT_BY_NAME ((uint)hintArray->u1.AddressOfData + idata->getAdjust()); 

			fprintf(stderr,"symbol: %s\n",a->Name);

			if (autoImportFlag)
				patch_address = (int *)&firstArray;
			else 
				patch_address = (int *)&firstArray->u1.Function; 
			
			fprintf(stderr,"patch_address %08x content %08x\n",patch_address, *patch_address);
				
  		char *name = (char *)a->Name; 

			uint addr = obj->exports->getVirtualAddress(name); 
			fprintf(stderr,"symaddr: %08x\n",addr + obj->ImageBase);
			*patch_address = (addr + obj->ImageBase); 				
		}
		// set 
		p->TimeDateStamp = 0xffffffff;
		p->ForwarderChain = 0xffffffff;
	}
	ntheader->FileHeader.TimeDateStamp = time(0);;

#if 1
	// fill bound import section 
	PIMAGE_DATA_DIRECTORY bdp = &ntheader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_BOUND_IMPORT]; 
	PIMAGE_SECTION_HEADER first_section = PIMAGE_SECTION_HEADER (ntheader+1);
	PIMAGE_BOUND_IMPORT_DESCRIPTOR bp_org = PIMAGE_BOUND_IMPORT_DESCRIPTOR (&first_section[ntheader->FileHeader.NumberOfSections]);
	PIMAGE_BOUND_IMPORT_DESCRIPTOR bp = bp_org; 
	char *bp2 = (char *)&bp[cache.getCount() + 1];
	
	cache.reset(); 

	while (obj = (LinkedObjectFile *)cache.getNext()) {
		bp->TimeDateStamp = time(0); 
		bp->OffsetModuleName = (uint)bp2 - (uint)bp_org;
//	bp->Reserved	
		bp->NumberOfModuleForwarderRefs = 0; 
		bp++; 

		strcpy(bp2,obj->getFileName());
		bp2 +=  strlen(obj->getFileName()) + 1; 
	}

	// set stop entry 	
	bp->TimeDateStamp = 0; 
	bp->OffsetModuleName = 0;
//	bp->Reserved	
	bp->NumberOfModuleForwarderRefs = 0; 

	// set data directory entry 
	bdp->VirtualAddress = (uint) bp_org - (uint)lpFileBase; 
	bdp->Size = (uint) bp2 - (uint) bp_org; 
#endif 
}	

LinkedObjectFile::~LinkedObjectFile() 
{
	if (exports)
			delete exports; 
	if (imports)
			delete imports; 
	if (relocs)
			delete relocs; 
}


ObjectFileList::ObjectFileList() 
{
	count = 0;
}

bool ObjectFileList::add(ObjectFile *obj)
{
	if (count < 1000) {
	 	list[count++] = obj;
		return true;
	}
	return false; 
}


ObjectFile *ObjectFileList::getNext(void) 
{
		if (iterator < count)
			return list[iterator++];
		else 
			return 0; 
}

ObjectFile *ObjectFileList::get(char *FileName)
{
	ObjectFile *p;
	reset(); 
	while (p = getNext()) {
		if (strstr(FileName,p->getFileName()))
			return p; 
	}		
	return 0; 
} 

ObjectFileList::~ObjectFileList() 
{
	ObjectFile *p;
	reset(); 
	while (p = getNext())
		delete p;
}


#ifdef OBJECTFILE_MAIN
int main(int argc, char **argv)
{
		ObjectFile test("/bin/cygz.dll");
		printf("%s\n",test.getFileName());


		LinkedObjectFile test("/bin/cygz.dll");

/* FIXME: update necessary 
		char *symbol = "gzgets"; 
		fprintf(stderr,"%-20s  %08x\n",symbol,test.getSymbolAddress(symbol));

		symbol = "_dist_code"; 
		fprintf(stderr,"%-20s  %08x\n",symbol,test.getSymbolAddress(symbol));

		symbol = "zlibVersion"; 
		fprintf(stderr,"%-20s  %08x\n",symbol,test.getSymbolAddress(symbol));
*/ 


		ObjectFileList test;
		test.Add("/bin/cygz.dll");
		ObjectFile *dll = test.getObjectFile("/bin/cygz.dll");
		if (dll) 
			printf("%s\n",dll->getFileName());


	return 0; 
}

#endif 
