/*
 * Copyright (c) 2002 Ralf Habacker  <Ralf.Habacker@freenet.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Id:$    
 */

#include <stdio.h>

#include "sections.h"

int Base::debug = 0;

Section::Section(void *aFileBase, PIMAGE_SECTION_HEADER p)
{
	header = p; 
	adjust = (uint)header->PointerToRawData + (uint) aFileBase - header->VirtualAddress; 
	strncpy(Name,(char *)header->Name,8);
	Name[8] = '\0';
}

void Section::print(char *title)
{
	fprintf(stderr,"%-10s name: %-8s: base: %08x size: %08x file offset: %08x offset: %08x\n",
			title,Name,header->VirtualAddress,header->SizeOfRawData,header->PointerToRawData, adjust);
}

bool Section::isIn(Section &in)
{
	return header->VirtualAddress >= in.header->VirtualAddress && header->VirtualAddress < in.header->VirtualAddress + in.header->SizeOfRawData;
}

bool Section::isIn(uint addr)
{
	return addr >= header->VirtualAddress && addr < header->VirtualAddress + header->SizeOfRawData;
}


SectionList::SectionList(void *aFileBase) 
{
	PIMAGE_DOS_HEADER dosheader = (PIMAGE_DOS_HEADER) aFileBase; 
	PIMAGE_NT_HEADERS ntheader = (PIMAGE_NT_HEADERS) ((char *)dosheader + dosheader->e_lfanew);

	header = (PIMAGE_SECTION_HEADER)	(ntheader+1);
	FileBase = (uint) aFileBase; 
	count = ntheader->FileHeader.NumberOfSections; 
	for (int i = 0; i < count; i++) {
		sections[i] = new Section(aFileBase,&header[i]);
		if (debug)
			sections[i]->print("section");
	}
}

SectionList::~SectionList() 
{
	for (int i = 0; i < count; i++)
		delete sections[i]; 
}

Section *SectionList::find(char *name)
{
	for (int i = 0; i < count; i++) {
		if ( strstr(sections[i]->getName(),name) )
			return sections[i];
	} 
	return 0; 
}

Section *SectionList::find(uint address)
{
	for (int i = 0; i < count; i++) {
		if (sections[i]->isIn(address))
			return sections[i];
	} 
	return 0; 
}


//----- class Exports --------------------------------------------------

Exports::Exports(Section &asection)
{
	adjust = asection.getAdjust(); 
	exports = PIMAGE_EXPORT_DIRECTORY (asection.getStartAddress()); 
}
	
Exports::Exports(SectionList &sections, PIMAGE_DATA_DIRECTORY iddp)
{
	Section *sec = sections.find(iddp->VirtualAddress); 
	if (sec) {
		adjust = sec->getAdjust(); 
		exports = (PIMAGE_EXPORT_DIRECTORY) (iddp->VirtualAddress + adjust);
	}
	else  {
		exports = 0; 
		adjust = 0; 
//		fprintf(stderr,"%s - error: can't find section for creating export object",__FUNCTION__);
	}
}

uint Exports::getVirtualAddress(char *symbol, uint *ordinal)
{
	if (!exports)
		return 0; 
		
	int n = exports->NumberOfFunctions; 
	uint *p = (unsigned int *)((char *)exports->AddressOfFunctions + adjust); 
	char **s = (char **)((char *)exports->AddressOfNames + adjust); 
	ushort *o = (ushort *)((char *)exports->AddressOfNameOrdinals + adjust); 
	for (int i = 0; i < n; i++,p++,s++) {
		if (strcmp(symbol,*s+adjust) == 0) {
			if (ordinal) 
				*ordinal = *o;
			return *p; 
		}
	}
	return 0; 
}

void Exports::reset(void)
{
		iterator =	0;
}
 
char *Exports::getNext(void) 
{
	if (!exports)
		return 0; 

	if (iterator < exports->NumberOfNames) {
		char **s = (char **)((char *)exports->AddressOfNames + adjust); 
		return (char *)(*(s+iterator++) + adjust); 
	}
	else 
		return 0;
}


void Exports::dump(char *title)
{

	char *p;
	printf("exports\n");
	if (!exports) 
		printf("\tno exports available\n");
	else {
		reset(); 
		while (p = getNext()) {
			printf("\t%s\n",p);
		}			
	}
}

//----- class Imports --------------------------------------------------


Imports::Imports(Section &asection)
{
	imports = PIMAGE_IMPORT_DESCRIPTOR (asection.getStartAddress()); 
	adjust = asection.getAdjust(); 
}

Imports::Imports(SectionList &sections, PIMAGE_DATA_DIRECTORY iddp)
{
	Section *sec = sections.find(iddp->VirtualAddress); 
	if (sec) {
		adjust = sec->getAdjust(); 
		imports = PIMAGE_IMPORT_DESCRIPTOR (iddp->VirtualAddress + adjust);
	}
	else  {
		imports = 0; 
		adjust = 0; 
//		fprintf(stderr,"%s - error: can't find section for creating import object\n",__FUNCTION__);
	}
}

void Imports::reset(void)
{
		iterator =	imports; 	
}
 
PIMAGE_IMPORT_DESCRIPTOR Imports::getNextDescriptor(void) 
{
	if (imports)
		return iterator->Name ? iterator++ : 0; 
	else 
		return 0; 
}

void Imports::dump(char *title)
{
	PIMAGE_IMPORT_DESCRIPTOR p;

	printf("imports\n");

	reset(); 
	while (p = getNextDescriptor()) {
		printf("%s \n",p->Name + adjust);
		printf("vma:           Hint     Time      Forward  DLL       First\n");
		printf("               Table    Stamp     Chain    Name      Thunk\n");
		printf("%08x       %08x %08x  %08x %08x  %08x\n",
			0,
			p->OriginalFirstThunk, 
			p->TimeDateStamp,
			p->ForwarderChain,
			p->Name,
			p->FirstThunk);
	}			
	printf("\n");
}

Relocations::Relocations(SectionList &sectionList, char *sectionName)
{
	sections = &sectionList; 
	Section *asection = sections->find(sectionName);
	if (asection) {
		relocs = PIMAGE_BASE_RELOCATION (asection->getStartAddress()); 
		adjust = asection->getAdjust(); 
		size = asection->getSize(); 
	}
	else {
		relocs = 0; 
		size = 0; 
	 }
}

bool Relocations::check(void)
{
	PIMAGE_BASE_RELOCATION relocp = relocs;
	Section *cursec; 
	int errors = 0; 

	if (!relocs)
		return false; 
	
	for (; (char *)relocp < (char *)relocs + size && relocp->SizeOfBlock != 0; ((char *)relocp) += relocp->SizeOfBlock  ) {
		int NumOfRelocs = (relocp->SizeOfBlock - sizeof(IMAGE_BASE_RELOCATION)) / sizeof (WORD); 
		int va = relocp->VirtualAddress; 
		PWORD p = (PWORD)((unsigned int )relocp + sizeof(IMAGE_BASE_RELOCATION));

		cursec = sections->find(va); 
		if (!cursec) {
			printf("warning: dll is corrupted - relocations for '%08x' are pointing to a non existent section\n",va);
			errors++; 
			continue; 
		}
	}		
	return errors == 0;
}
	
bool Relocations::fix(void)
{
	PIMAGE_BASE_RELOCATION relocp = relocs;
	Section *cursec; 
	int errors = 0; 

	if (!relocs)
		return false; 

	printf("warning: fixing bad relocations .... ");
	
	for (; (char *)relocp < (char *)relocs + size && relocp->SizeOfBlock != 0; ((char *)relocp) += relocp->SizeOfBlock  ) {
		int NumOfRelocs = (relocp->SizeOfBlock - sizeof(IMAGE_BASE_RELOCATION)) / sizeof (WORD); 
		int va = relocp->VirtualAddress; 
		PWORD p = (PWORD)((unsigned int )relocp + sizeof(IMAGE_BASE_RELOCATION));

		cursec = sections->find(va); 
		if (!cursec) {
			errors++; 
			relocp->SizeOfBlock = 0;
			relocp->VirtualAddress = 0;
			break;
		}
	}		
	if (errors == 0)
		printf("no errors found\n");
	else 
		printf("corrupted relocation records fixed\n");
	return true;
}


bool Relocations::relocate(int difference)
{
	PIMAGE_BASE_RELOCATION relocp = relocs;
	int WholeNumOfRelocs = 0;

	if (!relocs)
		return false; 
	
	for (; (char *)relocp < (char *)relocs + size && relocp->SizeOfBlock != 0; ((char *)relocp) += relocp->SizeOfBlock  ) {
		int NumOfRelocs = (relocp->SizeOfBlock - sizeof(IMAGE_BASE_RELOCATION)) / sizeof (WORD); 
		int va = relocp->VirtualAddress; 
		PWORD p = (PWORD)((unsigned int )relocp + sizeof(IMAGE_BASE_RELOCATION));
		if (debug) {
			printf("VirtAdress: %08x\n",relocp->VirtualAddress);
			printf("NumOfRelocs: %d\n",NumOfRelocs);
		}
		WholeNumOfRelocs += NumOfRelocs; 
	
		int adjust; 

		Section *cursec = sections->find(va); 
		if (!cursec) {
			printf("warning: dll is corrupted - the relocations '%08x' points to a non existing section and could not be relocated\n",va);
			continue; 
		}
		else 
			cursec->print("currently relocated section");		
		adjust = cursec->getAdjust(); 
			
		for (int i = 0; i < NumOfRelocs; i++,p++) {
			if ((*p & 0xf000) == 0x3000) {
				int location = (*p & 0x0fff) + va;
				if (debug) {
					fprintf(stderr,"%08x - ",location);
					fprintf(stderr,"%08x\n",location + adjust + 3);
				}
				int *patch_adr = (int *)cursec->rva2real(location); 
				*patch_adr +=	difference; 			
			}
		}
	}		
}	
	
